
%if 0

Assembly NASM port of INTERCEPT

intercep.c -- spawn a process and intercept BIOS and DOS
SWI (software interrupt) calls, recording register data to
a file.
Demonstrates chaining interrupt handlers.

 * Written 1987-07-31 by Ned Konz
 *
 * Released into the public domain by the author.
 *
 * Modified 1989-04-11 by Russell Nelson
 *
 * No copyright is claimed by Russell Nelson

2021 by C. Masloch

Parts taken from lDebug debugger,

Copyright (C) 1995-2003 Paul Vojta
Copyright (C) 2008-2021 C. Masloch

Usage of the works is permitted provided that this
instrument is retained with the works, so that any entity
that uses the works is notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.

%endif

	%assign __lMACROS1_MAC__DEBUG_DEFAULTS	1
%include "lmacros3.mac"
%include "AMIS.MAC"

	defaulting

	numdef PSPRELOC, 1		; relocate PSP

	numdef MAX_INTERRUPTS,	2500


	struc swi_info_v0
.bp:	resw 1
.di:	resw 1
.si:	resw 1
.ds:	resw 1
.es:	resw 1
.dx:	resw 1
.cx:	resw 1
.bx:	resw 1
.ax:	resw 1
.ip:	resw 1
.cs:	resw 1
.fl:	resw 1
.intnum:resw 1
	endstruc

	struc swi_info_v1_amis
.vendor:	resb 8
.product:	resb 8
.pointer:	resd 1
.version:	resw 1
.reserved:	resw 1
.flags:		resw 1
	endstruc
%if swi_info_v1_amis_size != swi_info_v0_size
 %error SWI info structures expected to have same size!
%endif


		; Option flags (hold in bp)
_outputfilename		equ 0001h
_datafilename		equ 0002h
_temppathname		equ 0004h
_filterfilename		equ 0008h
_longmode		equ 0010h
_amis			equ 0020h
_size			equ 0040h
_quiet			equ 0080h
_last_backslash		equ 0800h
_last_cr		equ 1000h
_questionmark		equ 2000h
_plus			equ 4000h
_gotcommand		equ 8000h

		; Formats (word format)
_format_raw	equ 8000h
_format_v0	equ 0
_format_v1	equ 1


%ifdef _MAP
	[map all _MAP]
%endif

	cpu 8086
	org 100h

	section TRANSIENT align=1	; Transient program part (command line parser, setup)

transient:
			; This is a simple MS-DOS 1.x detection (8 Byte).
	mov ah, 4Dh		; the only function I found that doesn't need much setup or cleanup
	stc			; but that uses MS-DOS 2.x error reporting. clears CF (Int21.30 does not)
	int 21h
	jnc .dosgood		; it is MS-DOS 2.x+ or compatible -->
	retn			; it isn't so simply terminate
.dosgood:
	cld

	mov dx, msg.logo
	call disp_msg

	mov word [outputfilename + 2], ds
	mov word [datafilename + 2], ds
	mov word [filterfilename + 2], ds
	mov word [temppathname + 2], ds
	mov word [progname + 2], ds
	mov word [execblock.fcb1 + 2], ds
	mov word [execblock.fcb2 + 2], ds
	mov word [linein], 255
	mov byte [linein + 2], 13

			; The UMB link and memory allocation strategy is now always safed here
			; and restored by abort_msg as well as any other termination code. This
			; saves any code changing the strategy the effort to restore it proper.
	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	mov word [ restored.umblink ], ax	; save UMB link

	mov ax, 5800h
	int 21h
	mov word [ restored.strategy ], ax	; save strategy

	mov dx, i23
	mov ax, 2523h
	int 21h			; set Int23 so that we restore link/strategy on Ctrl-C as well
	mov dx, i24
	mov ax, 2524h
	int 21h			; set Int24 so all hard errors are changed to the soft error

getfirstumcb:
			; We try to get the first UMCB for various uses later. Note that we
			; never use the result as an MCB address so if it's wrong we won't
			; crash or something. (We compare memory block or MCB addresses to
			; the value to check whether they're in the UMA.)
	mov ax, 1261h		; PTS-DOS: Get first UMCB
	stc
	int 2Fh
	jc .determine		; not supported -->
	inc ax
	cmp ax, byte 2		; -1, 0, 1 ?
	jbe .determine		; not supported (or none) -->
	dec ax
	mov word [ firstumcb ], ax	; set UMB
	jmp short .got		; got it -->

.determine:
	mov ax, 5803h
	xor bx, bx
	int 21h			; disable UMB link, leave only LMA chain
	jc .none		; that isn't supported either -->

	mov ah, 52h
	int 21h
	mov ax, word [ es:bx - 2 ]
	mov dx, ax		; first MCB
	xor bx, bx		; use offsets from bx, not addresses
.looplmb:
	mov ds, ax
	inc ax
	add ax, word [ bx + 3 ]	; next MCB's address
	cmp byte [ bx ], 'M'
	je .looplmb		; not last -->
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	xchg ax, dx		; dx = what we assume to be the first UMA chain MCB
				; ax = first MCB

	push ax
	inc bx			; = 1
	mov ax, 5803h
	int 21h			; enable UMB link, include UMA chain
	pop ax
	jc .none		; so we can disable it but not enable? -->

	dec bx			; = 0
	xor cx, cx		; flag if assumed first UMCB found
.loopumb:
	cmp ax, dx
	jne .notlastlmb
	inc cx			; there it is
.notlastlmb:
	mov ds, ax
	cmp byte [ bx ], 'M'
	jne .islastumb?		; last or corrupted -->
	inc ax
	add ax, word [ bx + 3 ]
	jmp short .loopumb	; process next -->
.islastumb?:
	cmp byte [ bx ], 'Z'
	jne .none		; corrupted -->
	jcxz .none		; didn't find that UMCB -->
			; The MCB at dx which was behind the one that contained the 'Z'
			; signature when we disabled the UMB link is now a valid MCB in
			; the MCB chain after we enabled the UMB link. All previous MCBs
			; are now 'M'.
	mov word [ cs:firstumcb ], dx
.none:
.got:
	call restorestate

	push cs
	pop ds
	mov ax, word [firstumcb]
	mov word [excludesegments.umcb], ax
	mov word [excludesegments.cs], cs

findtempenv:
	mov ax, word [2Ch]
	test ax, ax
	jz .done
	mov es, ax
	mov si, msg.tmpdireq
	mov cx, msg.tmpdireq.length
	call .searchvar
	jnc .got
	mov si, msg.tmpeq
	mov cx, msg.tmpeq.length
	call .searchvar
	jnc .got
	mov si, msg.tempeq
	mov cx, msg.tempeq.length
	call .searchvar
	jnc .got
	jmp .done

.searchvar:
	xor di, di
.searchloop:
	xor ax, ax
	push cx
	push si
	push di
	repe cmpsb
	je .foundvar
	pop di
	pop si
	mov cx, -1
	repne scasb
	jne .notfound
	scasb
	je .notfound
	dec di
	pop cx
	jmp .searchloop

.notfound:
	pop ax
	stc
	retn

.foundvar:
	pop ax
	pop ax
	pop ax
	clc
	retn

.got:
	mov word [temppathname], di
	mov word [temppathname + 2], es
.done:


findprogname:
	mov ax, word [2Ch]	; => environment
	test ax, ax
	jz .no_name
	mov ds, ax
	xor si, si
@@:
	lodsb
	test al, al
	jnz @B
	lodsb
	test al, al
	jnz @B
	lodsw
	cmp ax, 1
	jne .no_name
@@:
	mov bx, si
@@:
	lodsb
	cmp al, 'a'
	jb @F
	cmp al, 'z'
	ja @F
	sub byte [si - 1], 'a' - 'A'
@@:
	cmp al, '\'
	je @BBB
	cmp al, '/'
	je @BBB
	test al, al
	jnz @BB

	mov cx, si
	dec cx
	sub cx, bx

@@:
	dec si
	cmp si, bx
	jb @F
	cmp byte [si], '.'
	jne @B

	mov cx, si
	sub cx, bx
@@:
	jcxz .no_name
	push ds
	 push cs
	 pop ds
	pop word [progname + 2]
	mov word [progname], bx	; -> point to the found name
	mov word [progname.length], cx
				; = length of the found name
.no_name:


scancl:			; bp = selected options
	xor bp, bp		; initial option flags

	push cs
	pop es
	push cs
	pop ds			; ds, es changed while searching UMCB

		; Interpret switches and erase them from the command line.
		; From lDebug init.asm
	mov ax, 3700h			; get switch character
	mov dl, '/'			; preset with default value
	int 21h
	mov byte [ switchar ], dl
	cmp dl, '/'
	jne .notslash
	mov byte [ swch1 ], dl
.notslash:

	mov si, 81h		; PSP command line
.loop:	lodsb
.loop_have_al:
	cmp al, 9		; tab ?
	je .loop
	cmp al, 32		; blank ?
	je .loop
	jb .end			; zero, cr, lf ? -->
	cmp al, '-'
	je .switch
	cmp al, '/'
	je .switch
	call .setupexec
	jc .invalid
	or bp, _gotcommand
	jmp .end

.setupexec:
	mov bx, si
	dec bx
	db __TEST_IMM8
@@:
	lodsb
	cmp al, 32
	ja @B
	mov di, si
	dec di
	mov al, 0
	xchg al, byte [di]

	mov word [execname], bx
	mov word [execname + 2], ds

	mov dh, al		; dh = first byte of tail
	mov bx, si
	dec bx			; bx -> start of tail
	db __TEST_IMM8
@@:
	lodsb
	cmp al, 13
	jne @B			; si -> after 13 terminator

	mov di, si
	dec si			; si -> at 13 terminator
	inc di			; di = si + 2
	mov cx, si
	sub cx, bx		; length of tail excluding terminator
	mov dl, cl		; dl = length of tail excluding terminator
	inc cx			; copy tail including terminator
	std			; AMD erratum 109 workaround follows
	cmp cx, 20
	ja @FF
	jcxz @FF
@@:
	movsb
	loop @B
@@:
	rep movsb
	cld
	inc bx			; -> tail structure
	mov word [bx], dx	; set tail length and tail first byte

	mov word [execblock.tail], bx
	mov word [execblock.tail + 2], ds

	mov al, 0
	mov di, word [execname]
	scasb
	je .setupexec.invalid

	mov si, bx
	inc si

.fcb_setup:
		; Set up FCBs.
	mov di, 5Ch
	call nn6		; do first FCB
	mov di, 6Ch
	call nn6		; second FCB

	db __TEST_IMM8		; (skip stc, NC)
.setupexec.invalid:
	stc
	retn


			; parse switch
.switch:
	lodsb			; get next character
	call capitalise
			; check against known switches
	cmp al, 'X'
	je .multiplexnumber
	cmp al, 'F'
	je .format
	cmp al, 'E'
	je .exclude
	push si
	mov si, parameters
	xchg al, ah
.loopswitch:
	lodsb
	xchg ax, cx
	lodsb
	xchg ax, cx		; cx = option flags
	cmp al, 0FFh
	je .invalidswitch	; not found!
	cmp al, ah
	jne .loopswitch		; try next -->
	pop si
	xor ch, ch
	or bp, cx
	jcxz .help
	test cl, _size
	jnz .parametersize
	test cl, _outputfilename | _datafilename | _temppathname | _filterfilename
	jnz .pathname
	jmp .loop		; ds:si-> character after switch character. just continue


.invalidswitch:
	call disp_progname
	mov dx, msg.invalidswitch.1
	call disp_msg
	call disp_al
	mov dx, msg.invalidswitch.2
	jmp abort_msg


.help:
	mov dx, msg.help.1
	call disp_msg
	call disp_progname
	mov dx, msg.help.2
	call disp_msg
	jmp ..@exit


.invalid:
	call disp_progname
	mov dx, msg.invalidparameter
	jmp abort_msg


.pathname:
	lodsb
	mov bx, si
	cmp al, '='
	je @FF
	cmp al, ':'
	je @FF

	db __TEST_IMM8
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	jb .invalid

	mov bx, si
	dec bx
	db __TEST_IMM8
@@:
	lodsb
	cmp al, 32
	ja @B
	mov di, si
	dec di
	cmp di, bx
	je .invalid
	mov al, 0
	xchg al, byte [di]
	mov di, outputfilename
	test cl, _outputfilename
	jnz @F
	mov di, datafilename
	test cl, _datafilename
	jnz @F
	mov di, temppathname
	test cl, _temppathname
	jnz @F
	mov di, filterfilename
	test cl, _filterfilename
	jnz @F
	jmp .invalid

@@:
	mov word [di], bx
	mov word [di + 2], ds
	jmp .loop_have_al


.parametersize:
	lodsb
	cmp al, '='
	je @FF
	cmp al, ':'
	je @FF

	db __TEST_IMM8
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	jb .invalid

	db __TEST_IMM8
@@:
	lodsb
	call getdecit
	jc .invalid
	xor bx, bx
	mov bl, al
@@:
	lodsb
	call getdecit
	jc .check_end_size_number
	mov cx, bx
	add bx, bx		; * 2
	jc .invalid
	add bx, bx		; * 4
	jc .invalid
	add bx, cx		; * 5
	jc .invalid
	add bx, bx		; * 10
	jc .invalid
	add bl, al
	adc bh, 0
	jc .invalid
	jmp @B
.check_end_size_number:
	cmp al, 9
	je @F
	cmp al, 32
	ja .invalid
@@:

	dec si
	mov word [size], bx
	jmp .loop


.multiplexnumber:
	lodsb
	cmp al, '='
	je @FF
	cmp al, ':'
	je @FF

	db __TEST_IMM8
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	jb .invalid

	db __TEST_IMM8
@@:
	lodsb
	call getnybble
	jc .invalid
	xor bx, bx
	mov bl, al
	lodsb
	call getnybble
	jc .check_end_multiplexnumber
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx
	add bl, al
	lodsb
.check_end_multiplexnumber:
	cmp al, 9
	je @F
	cmp al, 32
	ja .invalid
@@:

	xchg bh, bl
	mov word [trymultiplexnumber], bx
	dec si
.loop_1:
	jmp .loop


.format:
	lodsb
	cmp al, '='
	je @FF
	cmp al, ':'
	je @FF

	db __TEST_IMM8
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	jb .invalid

	dec si
@@:
	mov dx, msg.raw
	call isstring?
	jne @F
	setopt [format], _format_raw
	jmp .loop_1

@@:
	mov dx, msg.v0raw
	call isstring?
	jne @F
	mov word [format], _format_raw | _format_v0
	jmp .loop_1

@@:
	mov dx, msg.v0
	call isstring?
	jne @F
	mov word [format], _format_v0
	jmp .loop_1

@@:
	mov dx, msg.v1raw
	call isstring?
	jne @F
	mov word [format], _format_raw | _format_v1
	jmp .loop_1

@@:
	mov dx, msg.v1
	call isstring?
	jne @F
	mov word [format], _format_v1
	jmp .loop_1

@@:
	jmp .invalid


.exclude:
	lodsb
	cmp al, '='
	je @FF
	cmp al, ':'
	je @FF

	db __TEST_IMM8
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	jb .invalid

	dec si
@@:
	mov dx, msg.none
	call isstring?
	jne @F
	and word [excludesegments.inuse], 0
	jmp .loop_1

@@:
	jmp .invalid


.end:
	test bp, _gotcommand
	jz .help


findinstalleddebugger:
	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	call .check
	mov al, 30h		; al = 30h to indicate found, ah = multiplex number
	jnc .end
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't check all yet -->

	xor ax, ax		; al = 0 to indicate none found
	jmp .end		;  If not found, continue -->


		; INP:	ah = multiplex number to check
		;	ds = ss = cs
		; OUT:	CY if multiplex number unused or no signature match,
		;	 bp, ah, ds unmodified
		;	NC if match found,
		;	 ah = multiplex number (unmodified)
		; CHG:	si, di, es, cx, dx
.check:
	mov al, 00h		; AMIS installation check
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
	cmp al, 0FFh
	jne .notfound
	mov si, msg.debuggeramissig
				; ds:si -> our AMIS name strings
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
	mov cx, 8		; Ignore description, only compare vendor and program name
	repe cmpsw
	je .checkret		; ZR, NC = match -->
.notfound:
	stc			; NZ, CY no match
.checkret:
	retn

.end:
	mov word [debuggerfunction], ax


process:
	mov word [options], bp

	cmp sp, stacktop_with_buffer
	jae @F

	call disp_progname
	mov dx, msg.error_memory
	jmp abort_msg

@@:
	push cs
	pop es
	mov sp, stacktop_with_buffer
	mov ah, 4Ah
	mov bx, (stacktop_with_buffer - buffers + transient_size + 256 + 15) >> 4
	int 21h
	mov dx, msg.internalerror
	jc abort_msg

readdata:
	lds dx, [datafilename]
	mov ax, 3D00h + 1_010_0_000b	; open, do not inherit, DENY_WRITE, RO
	int 21h
	jnc @F

.erroropening:
	call disp_progname
	mov dx, msg.error_opening.1
	call disp_msg
	mov di, datafilename
	call disp_filename
	mov dx, msg.error_opening.2
	call disp_dos_error
	jmp abort

.errorreading:
	call disp_progname
	mov dx, msg.error_reading.1
	call disp_msg
	mov di, datafilename
	call disp_filename
	mov dx, msg.error_reading.2
	call disp_dos_error
	jmp abort

@@:
	push cs
	pop ds
	mov word [handle], ax
	mov di, .loop

		; This is a little state machine.
		; di -> current state handler
		; cx = bytes left + 1 (ie 1 = no bytes, 2 = one byte, etc)
		; ds:si -> next (cx - 1) bytes in buffer
		; bx is preserved across .outerloop and can be used to
		;  hold additional state (generally a number being parsed)
		; other additional state has to be held in memory
		; each state handler starts with a mov to di and then
		;  a loop instruction which jumps to the further handler.
		;  if the loop instruction is executed with cx = 1 it
		;  does not branch; after that there's a jump back to
		;  .outerloop which will try reading more data.
		; if the template file data is exhausted we either jump
		;  to .end or we fill the buffer with a CR, cx = 2
		;  (one byte available) and jump to the state handler.
.outerloop:
	push bx
	mov dx, buffer
	mov cx, buffer_size
	mov bx, word [handle]
	mov ah, 3Fh
	int 21h
	pop bx
	jc .errorreading
	test ax, ax
	jnz @F

	cmp di, .letmefinish.start
	jb .end
	cmp di, .letmefinish.end
	jae .end
	mov cx, 2
	mov si, dx
	mov byte [si], 13
	jmp di

@@:
	xchg cx, ax

	mov si, dx
	inc cx
	jmp di

		; Default start state: Skip blanks.
.loop:
	mov di, .loop
	loop @F
	jmp .outerloop

@@:
	lodsb

	test bp, _last_cr
	jz @F
	cmp al, 13		; another CR ?
	je .next_al_clear_backslash
				; handle as next -->
	and bp, ~ _last_cr
	cmp al, 10		; is an LF after CR ?
	je .loop		; yes, just loop (without incrementing line number) -->
@@:

	test bp, _last_backslash
	jz @F
.next_al_clear_backslash:
	dec si
	inc cx			; unget the non-EOL byte
	jmp .next_clear_backslash

@@:
	cmp al, 32		; blank ?
	je .loop
	cmp al, 9
	je .loop		; skip it -->

	cmp al, '+'		; plus sign ?
	jne .checknybble

		; Plus detected, skip blanks again
.plus:
	or bp, _plus		; remember plus is in effect
	mov di, .plus
	loop @F
	jmp .outerloop

@@:
	lodsb
	cmp al, 32
	je .plus
	cmp al, 9
	je .plus
	cmp al, '?'		; question mark ?
	jne .checknybble	; check for hex digit else -->

		; Question mark detected after plus, skip blanks again
.questionmark:
	or bp, _questionmark	; remember question mark is in effect
	mov di, .questionmark
	loop @F
	jmp .outerloop

@@:
	lodsb
	cmp al, 32
	je .questionmark
	cmp al, 9
	je .questionmark

		; Check for an interrupt number,
		;  INP:al = loaded byte
.checknybble:
	call getnybble		; is it a hex digit ?
	jc .next_al		; no, skip to end of line -->

	xor bx, bx
	mov bl, al		; save number

		; Below states need finish handling
		;  at EOF to complete their processing.
.letmefinish.start:
		; Check for second digit
		; bx = number read from first digit
.checkseconddigit:
	mov di, .checkseconddigit
	loop @F
	jmp .outerloop

@@:
	lodsb			; load candidate second digit
	call getnybble
	jc .gotnumber
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx		; multiply first digit by 16
	add bl, al		; add in second digit
	db __TEST_IMM16		; second digit consumed
.gotnumber:
	dec si			; unget current byte
	inc cx

	mov ax, bp		; are we in template to filter out as quiet ?
	and ax, (_quiet | _questionmark)
	cmp ax, (_quiet | _questionmark)
	jne .nonquietnumber	; no -->

	mov ax, word [excludeints.inuse]
				; ax = amount used
	cmp ax, word [excludeints.amount]
				; all used already ?
	jb @F			; if not -->

	call disp_progname
	mov dx, msg.templatetoomanyquiet.1
	call disp_msg
	mov di, datafilename
	call disp_filename
	mov dx, msg.templatetoomanyquiet.2
	call disp_msg
	mov ax, word [linenumber]
	call disp_ax_dec
	mov dx, msg.templatetoomanyquiet.3
	jmp abort_msg

.invalid:
	call disp_progname
	mov dx, msg.templateinvalid.1
	call disp_msg
	mov di, datafilename
	call disp_filename
	mov dx, msg.templateinvalid.2
	call disp_msg
	mov ax, word [linenumber]
	call disp_ax_dec
	mov dx, msg.templateinvalid.3
	jmp abort_msg

@@:
	mov di, ax
	add di, di
	add di, di		; = byte offset to our next exclude sttucture
	mov byte [excludeints.first + di + 0], bl
				; store our interrupt number

		; Check for ah value first digit
		;
		; At this point a blank or EOL byte is
		;  valid to end the template without
		;  any numeric value for AH (or AL).
		; Also valid is 'M' for the special
		;  wildcard specifier (for int 2Dh),
		;  which just means "do not match AH
		;  but allow to match AL still".
		; If neither of the above two cases is
		;  true then the next byte *MUST* be
		;  a hex digit, which we check here.
.checkfirstdigitah:
	mov di, .checkfirstdigitah
	loop @F
	jmp .outerloop

@@:
	lodsb			; load byte
	call capitalise
	cmp al, 9		; blank or EOL ?
	je .finishquiet
	cmp al, 32
	je .finishquiet
	cmp al, 13
	je .finishquiet
	cmp al, 10
	je .finishquiet		; yes, finish quiet filtered template -->
	cmp al, 'M'		; M special wildcard for ah ?
	je .checkseconddigitmah	; yes -->
	call getnybble		; else must be a valid hex digit
	jc .invalid		; if not -->

	xor bx, bx
	mov bl, al		; remember first digit for ah

		; Check for ah value second digit
		;
		; This must be a hex digit.
.checkseconddigitah:
	mov di, .checkseconddigitah
	loop @F
	jmp .outerloop

@@:
	lodsb			; load byte
	call getnybble		; is a hex digit ?
	jc .invalid		; no -->
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx		; first digit times 16
	add bl, al		; add in second digit

	mov di, word [excludeints.inuse]
	add di, di
	add di, di		; byte index to current exclude structure
	mov byte [excludeints.first + di + 3], bl
				; remember the number
	or byte [excludeints.first + di + 1], eifMatchAH
				; remember to match AH

		; Check for al first digit
		;
		; Blank or EOL is valid here to indicate
		;  no value for AL, else there must be a
		;  hex digit pair here.
.checkfirstdigital:
	mov di, .checkfirstdigital
	loop @F
	jmp .outerloop

@@:
	lodsb			; load byte
	cmp al, 9		; blank ?
	je .finishquiet
	cmp al, 32
	je .finishquiet
	cmp al, 13
	je .finishquiet
	cmp al, 10
	je .finishquiet		; yes -->
	call getnybble		; hex digit ?
	jc .invalid		; no, invalid -->

	xor bx, bx
	mov bl, al		; remember first digit

		; Check for al second digit
		;
		; This must be a hex digit.
.checkseconddigital:
	mov di, .checkseconddigital
	loop @F
	jmp .outerloop

@@:
	lodsb			; load byte
	call getnybble		; get hex digit
	jc .invalid		; else error -->
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx		; first digit times 16
	add bl, al		; add in second digit

	mov di, word [excludeints.inuse]
	add di, di
	add di, di		; byte offset to current exclude structure
	mov byte [excludeints.first + di + 2], bl
				; remember value to match AL
	or byte [excludeints.first + di + 1], eifMatchAL
				; remember to match AL
	jmp .finishquiet.next

		; Check that the next byte is the second 'M'
.checkseconddigitmah:
	mov di, .checkseconddigitmah
	loop @F
	jmp .outerloop

@@:
	lodsb			; load byte
	call capitalise		; capitalise letters
	cmp al, 'M'		; is it M ?
	jne .invalid		; no -->

	jmp .checkfirstdigital	; allow optional AL match number next -->


.letmefinish.end:

.finishquiet:
	dec si			; unget a byte
	inc cx
.finishquiet.next:		; entry for having consumed the byte
	mov di, word [excludeints.inuse]
	add di, di
	add di, di		; get byte index to current exclude structure
	inc word [excludeints.inuse]
				; allocate the structure now
	; mov bl, byte [excludeints.first + di + 0]
				; restore interrupt number
	jmp .next		; do NOT hook the interrupt if quiet
				;  (this only makes a difference if there
				;  are only quiet templates for a given
				;  interrupt, ie only int 28h and 29h.)

.nonquietnumber:
	mov al, bl		; copy interrupt number to al

	 push cx
	call bx_intnum_to_dh_bx_intsel
				; get mask and offset for bit array
	push cs
	pop es
	mov di, invalidinterrupts
	mov cx, invalidinterrupts_size
				; is it a valid interrupt ?
	repne scasb		; scan invalid list
	 pop cx
	jne @F			; if is valid -->

	call disp_progname
	mov dx, msg.invalid_interrupt_1
	call disp_msg
	call disp_al_hex
	mov dx, msg.invalid_interrupt_2
	jmp abort_msg

@@:
	or byte [selectedinterrupts + bx], dh
				; mark this interrupt as hooking

		; Done with this template, advance to next line
.next:
	and bp, ~ (_plus | _questionmark)
				; clear these flags
	mov di, .next
	loop @F
	jmp .outerloop

@@:
	lodsb			; get byte
.next_al:			; entry for byte already loaded
	cmp al, '\'
	jne @F
	or bp, _last_backslash
	jmp .next

@@:
	cmp al, 13		; EOL ?
	je .13
	cmp al, 10
	je .10			; yes, return to start state -->
.next_clear_backslash:
	and bp, ~ _last_backslash
	jmp .next		; continue to skip -->

.13:
	or bp, _last_cr
.10:
	inc word [linenumber]
	jmp .loop


.end:
	mov bx, word [handle]
	mov ah, 3Eh
	int 21h

	call disp_progname
	mov dx, msg.templatedone.1
	call disp_msg
	mov di, datafilename
	call disp_filename
	mov dx, msg.templatedone.2
	call disp_msg
	mov ax, word [linenumber]
	call disp_ax_dec
	mov dx, msg.templatedone.3
	call disp_msg


remove_not_hooked_quiet:
	push cs
	pop es
	mov si, excludeints.first
				; si -> exclude structures
	mov cx, word [excludeints.inuse]
				; amount
.loop_or_zero:
	jcxz .done
.loop:
	xor bx, bx
	mov bl, [si]		; interrupt for this quiet template
	push cx
	call bx_intnum_to_dh_bx_intsel
				; get mask and offset for bit array
	pop cx
	test byte [selectedinterrupts + bx], dh
	jnz .next		; skip to next -->

	push si
	mov di, si		; -> entry to overwrite
	add si, 4		; -> entry after that
	dec cx			; = count of entries after that
	push cx
	shl cx, 1		; = count of words
	rep movsw		; move down
	pop cx			; = count of remaining entries (may be zero)
	pop si			; -> next structure to process
	dec word [excludeints.inuse]
	jmp .loop_or_zero

.next:
	add si, 4		; -> next structure
	loop .loop		; count down
.done:


mark_quiet:
	mov si, excludeints.first
				; si -> exclude structures
	mov cx, word [excludeints.inuse]
				; amount
	jcxz .done
.loop:
	xor bx, bx
	mov bl, [si]		; interrupt for this quiet template
	push cx
	call bx_intnum_to_dh_bx_intsel
				; get mask and offset for bit array
	pop cx
	or byte [quietinterrupts + bx], dh
	add si, 4		; -> next structure
	loop .loop		; count down
.done:


initctrl:
	test bp, _amis
	jz @F
	or byte [ctrl0], 40h

@@:

allocateinfo:
	mov sp, stacktop_without_buffer
	mov ah, 4Ah
	mov bx, (stacktop_without_buffer - buffers + transient_size + 256 + 15) >> 4
	int 21h
	mov dx, msg.internalerror
	jc abort_msg

	mov ax, word [size]
	mov bx, swi_info_v0_size
	mul bx
	add ax, 15
	adc dx, 0
	jc .error

	mov cx, 4
@@:
	shr dx, 1
	rcr ax, 1
	loop @B

	test dx, dx
	jnz .error

	mov bx, ax
	mov ah, 48h
	int 21h
	jc .error
	mov word [swi_info_array_segment], ax
	jmp .end

.error:
	call disp_progname
	mov dx, msg.error_memory
	jmp abort_msg

.end:


allocatehook:
	xor di, di
	xor ax, ax
.loop:
	mov bx, ax
	cmp al, 2Dh
	je .next
	call bx_intnum_to_dh_bx_intsel
	test byte [selectedinterrupts + bx], dh
	jz .next
	inc di
.next:
	inc al
	jnz .loop

	mov word [int_amount], di

	mov ax, handler_size + 3	; handler + int list slot size
	mul di

	test dx, dx
	jnz .error

	add ax, 3 + 1			; list slot for 2Dh + round up
	jc .error
	and al, ~1			; round to even boundsry

	add ax, stub_size + 15		; stub size + round up paragraph
	jc .error

	mov di, ax
	and di, ~15			; -> behind end of block
	mov cx, 4
@@:
	shr dx, 1
	rcr ax, 1
	loop @B

	test dx, dx
	jnz .error

	mov bx, ax
	mov ah, 48h
	int 21h
	jc .error
	mov word [int_segment], ax

.error: equ allocateinfo.error


inithandlers:
	mov es, word [int_segment]
	xor ax, ax
	mov cx, 8
	sub di, 16		; -> last paragraph
	rep stosw		; initialise it

	mov ax, word [trymultiplexnumber]
	test al, al		; set ?
	jnz @F			; no -->
	int 2Dh
	test al, al		; still 00h if free
	jz .foundamis
@@:

	mov ah, 00h		; start with multiplex number 00h
.loopamis:
	mov al, 00h		; installation check
	int 2Dh
	test al, al		; still 00h if free
	jz .foundamis
	inc ah			; search from bottom to top
	jnz .loopamis		; loop if zero not yet reached -->
	call disp_progname
	mov dx, msg.noamisnumber; abort if no free AMIS multiplex number left
	jmp abort_msg

.foundamis:
			; ah = free AMIS multiplex number
	mov byte [i2D.multiplexnumber], ah
				; set multiplex number

	mov word [int2Dhandler.cs], cs

	xor di, di
	mov si, stub
	mov word [stub.segment1], cs
	mov word [stub.segment2], cs
	mov cx, stub_size_w
	rep movsw

	mov bx, di		; -> start of int list
	mov word [amisintr], di
	mov ax, word [int_amount]
	inc ax			; count entry for 2Dh handler
	add di, ax
	add di, ax
	add di, ax		; -> behind last list slot
	inc di			; to round up
	and di, ~1		; round to even boundary

	dec di
	dec di			; -> last word
	xor ax, ax		; = 0
	stosw			; store into the last list word to initialise it
.loop:
	cmp al, 2Dh		; is 2Dh ?
	je .next		; yes, handle specifically -->

	push bx
	mov bx, ax		; is it selected ?
	call bx_intnum_to_dh_bx_intsel
	test byte [selectedinterrupts + bx], dh
	pop bx
	jz .next		; no -->

	mov byte [es:bx], al	; store interrupt number
	inc bx
	mov word [es:bx], di	; store offset of handler
	inc bx
	inc bx			; -> next int list slot
	mov byte [handler.number], al
				; store number in our handler
	lea dx, [di + handler.displacebase - handler]
				; = base of displacement (behind E8h near call)
	neg dx			; = minus base
	add dx, stub.entry - stub
				; = entry minus base
	mov word [handler.displacement], dx
				; patch in the displacement

	mov si, handler		; -> prepared handler
	mov cx, handler_size_w
	rep movsw		; write to memory

.next:
	inc al
	jnz .loop

	mov byte [es:bx], 2Dh	; last int list slot is 2Dh
	inc bx
	mov word [es:bx], i2D - stub
				; relocated i2D handler
	mov word [int_end], di

	testopt [format], _format_raw
	jnz .prepare_no_filter

	xor dx, dx
	les di, [temppathname]
	call strlen
	add cx, 14
	jc .error
	add dx, cx
	jc .error
	add dx, cx
	jc .error

	les di, [datafilename]
	call strlen
	add dx, cx
	jc .error

	les di, [outputfilename]
	call strlen
	add dx, cx
	jc .error

	les di, [execname]
	call strlen
	add dx, cx
	jc .error

	add dx, 16 + 8 + 4
	jc .error

	cmp dx, word [int_end]
	ja .error
.prepare_no_filter:

.error: equ allocateinfo.error


displayrecording:
	call disp_progname
	mov dx, msg.recording.1
	call disp_msg
	mov ax, word [size]
	call disp_ax_dec
	mov dx, msg.recording.2
	call disp_msg
	mov di, outputfilename
	call disp_filename
	mov dx, msg.recording.3.raw
	testopt [format], _format_raw
	jnz @F
	mov dx, msg.recording.3.notraw
@@:
	call disp_msg


installhandlers:
	call disp_progname
	mov dx, msg.hooking.1
	call disp_msg
	mov ax, word [int_amount]
	call disp_ax_dec
	mov dx, msg.hooking.2
	call disp_msg
	mov al, byte [i2D.multiplexnumber]
	call disp_amisnum

	lds bx, [amisintr]	; ds:bx -> interrupt list of image
	mov dx, HookInterruptSim
	mov cx, .simulated
	push ds
	push bx
	jmp short .loopamisintr	; do the hook simulation
.simulated:
	jc hookerror		; failed -->
	pop bx
	pop ds

			; (Note that in case of critical hook failure, we still want the
			; handler to stay allocated in case some of the already hooked
			; interrupts can't be unhooked anymore. Otherwise, the code to
			; handle a nullified critical failure will free the MCB itself.)

	; mov bx, [cs:amisintr]	; ds:bx -> AMIS interrupt list (bx still valid)
	mov dx, HookInterrupt
	mov cx, .hooked
.loopamisintr:
	jmp loopamisintr	; really hook all interrupts in list
.hooked:
	jc handle_crithookfailure


enable:
	setopt [cs:ctrl0], 80h

execute:
	push cs
	pop ds
	push cs
	pop es
	mov bx, execblock
	lds dx, [execname]
	mov ax, 4B00h
	int 21h


disable:
	pushf
	clropt [cs:ctrl0], 80h
	popf

epilogue:
	jnc @F
	call disp_progname
	mov dx, msg.exec_error_user
	call disp_dos_error
	stc
	jmp @FF

@@:
	mov ah, 4Dh
	int 21h
	call disp_progname
	mov dx, msg.terminated.1
	call disp_msg
	call disp_ax_hex
	mov dx, msg.terminated.2
	call disp_msg
	call disp_ax_dec
	mov dx, msg.terminated.3
	call disp_msg
	clc
@@:


uninstallhandlers:
	pushf

	lds bx, [cs:amisintr]	; ds:bx -> interrupt list of image
	push ds
	push bx
	mov dx, UnhookInterruptSim
	mov cx, .simulated
	jmp short .loopamisintr
.simulated:
	jc unhookerror
	pop bx
	pop ds			; restore ds:bx ->

	; mov bx, [cs:amisintr]	; ds:bx -> AMIS interrupt list (bx still valid)
	mov dx, UnhookInterrupt
	mov cx, .unhooked
.loopamisintr:
	jmp loopamisintr
.unhooked:
	mov cx, panic_and_critunins
	jc unhookerrorcritical		; (cx non-zero: display "critical error")

	popf

abort_if_exec_error:
	jc abort


createtemp:
	push cs
	pop ds
	call disp_progname
	mov dx, msg.usedtotal.1
	call disp_msg
	mov ax, word [free]
	call disp_ax_dec
	mov dx, msg.usedtotal.2
	call disp_msg
	mov ax, word [size]
	call disp_ax_dec
	mov dx, msg.usedtotal.3
	call disp_msg

	testopt [format], _format_raw
	jnz .raw

	xor dx, dx
	les di, [temppathname]
	call strlen
	jcxz .appendsep
	cmp byte [es:di - 2], '\'
	je .gotsep
	cmp byte [es:di - 2], '/'
	je .gotsep
.appendsep:
	inc dx
.gotsep:
	mov es, word [int_segment]
	lds si, [temppathname]
	xor di, di
	rep movsb
	test dx, dx
	jz @F
	mov byte [es:di], '\'
	inc di
@@:
	mov cx, 13
	rep stosb

	mov word [cs:aftertempfilename], di

	push es
	pop ds
	xor dx, dx			; -> temp pathname, '\', 13 NULs
	xor cx, cx			; file attribute to create
	mov ah, 5Ah
	int 21h				; create temporary file
	jnc @F

	call disp_progname
	mov dx, msg.error_create.1
	call disp_msg
	mov di, temppathname
	call disp_filename
	mov dx, msg.error_create.2
	jmp abort_msg

.raw:
	lds dx, [outputfilename]
	xor cx, cx
	mov ah, 3Ch
	int 21h
	jnc @F

	call disp_progname
	mov dx, msg.error_create_raw.1
	call disp_msg
	mov di, outputfilename
	call disp_filename
	mov dx, msg.error_create_raw.2
	jmp abort_msg

@@:
	push cs
	pop ds
	mov word [handle], ax
	xchg bx, ax


writetemp:
	mov di, word [free]
	mov ds, word [swi_info_array_segment]
	mov ax, swi_info_v0_size
	mul di				; dx:ax = number of bytes to write
	mov di, ax
	mov si, dx			; si:di = amount bytes
.loop:
	sub di, 8192
	sbb si, 0			; count down
	jc .last			; no full block left -->
	xor dx, dx			; ds:dx -> data to write
	mov cx, 8192			; write full block
	mov ah, 40h
	int 21h				; write to file
	jc .error
	cmp cx, ax
	jne .error_full

.next:
	mov ax, ds
	add ax, 8192 >> 4
	mov ds, ax			; => next block after full block
	jmp .loop

.last:
	add di, 8192			; restore di
	; adc si, 0			; (would be zero, not needed)
	jz .done			; if data was block-aligned -->
	xor dx, dx
	mov cx, di			; write partial block
	mov ah, 40h
	int 21h				; write to file
	jc .error
	cmp cx, ax
	jne .error_full
	jmp .done

.error_full:
	mov ax, 39			; "insufficient disk space" error
.error:
	push cs
	pop ds

	call disp_progname
	mov dx, msg.writeerror.1
	call disp_msg
	mov di, outputfilename
	testopt [format], _format_raw
	jnz @F
	mov di, amisintr
	and word [di], 0
@@:
	call disp_filename
	mov dx, msg.writeerror.2
	call disp_dos_error

	mov bx, [handle]
	mov ah, 3Eh
	int 21h

	mov di, outputfilename
	testopt [format], _format_raw
	jnz @F
	mov di, amisintr
@@:
	mov ah, 41h
	lds dx, [di]
	int 21h

	jmp abort

.done:
	push cs
	pop ds

	mov bx, [handle]
	mov ah, 3Eh
	int 21h

	testopt [format], _format_raw
	jnz ..@exit


callfilter:
	mov es, word [int_segment]
	mov di, word [aftertempfilename]
	mov ax, 32 << 8
	stosw
	mov dx, word [format]
	mov ax, "v1"
	cmp dx, _format_v1
	jb @F
	; je .gotformat

.gotformat:
	push ax
	mov ax, "-f"
	stosw
	mov al, 32
	stosb
	pop ax
	stosw
	mov al, 32
	stosb
@@:
	test bp, _longmode
	jz @F
	mov ax, "-l"
	stosw
	mov al, 32
	stosb
@@:
	test bp, _quiet
	jz @F
	mov ax, "-q"
	stosw
	mov al, 32
	stosb
@@:
	mov ax, "-t"
	stosw
	mov al, 32
	stosb
	mov dx, di

	les di, [datafilename]
	call strlen
	mov es, word [int_segment]
	mov di, dx
	lds si, [datafilename]
	rep movsb
	push cs
	pop ds

	mov al, 32
	stosb
	mov dx, di

	mov di, amisintr
	and word [di], 0
	les di, [amisintr]
	call strlen
	; mov es, word [int_segment]
	mov di, dx
	lds si, [amisintr]
	rep movsb
	push cs
	pop ds

	mov al, 32
	stosb
	mov dx, di
	les di, [outputfilename]
	call strlen
	mov es, word [int_segment]
	mov di, dx
	lds si, [outputfilename]
	rep movsb
	push cs
	pop ds

	mov al, 32
	stosb
	mov dx, di
	les di, [execname]
	call strlen
	mov es, word [int_segment]
	mov di, dx
	lds si, [execname]
	rep movsb
	push cs
	pop ds

	mov al, 13
	stosb
	dec di
	mov si, word [aftertempfilename]
	inc si
	mov cx, di
	sub cx, si
	jc .errortoolong

	cmp cx, 126
	ja .errortoolong

	dec si
	mov byte [es:si], cl

	push cs
	pop ds

	mov word [execblock.tail], si
	mov word [execblock.tail + 2], es

	push cs
	pop es

	mov bx, execblock
	lds dx, [filterfilename]
	mov ax, 4B00h
	int 21h

	jnc @F
	call disp_progname
	mov dx, msg.exec_error_filter
	call disp_dos_error

	jmp .couldnt

.errortoolong:
	call disp_progname
	mov dx, msg.error_cmdtail_too_long
	call disp_msg

	jmp .couldnt


@@:
	push cs
	pop ds
	mov ah, 4Dh
	int 21h

	test ax, ax
	jz @F

.error:
	call disp_progname
	mov dx, msg.filterreturn.1
	call disp_msg
	call disp_ax_hex
	mov dx, msg.filterreturn.2
	call disp_msg
	call disp_ax_dec
	mov dx, msg.filterreturn.3
	call disp_msg

.couldnt:
	mov ah, 41h
	lds dx, [amisintr]
	int 21h

	call disp_progname
	mov dx, msg.couldntfilter.1
	call disp_msg
	mov di, filterfilename
	call disp_filename
	mov dx, msg.couldntfilter.2
	jmp abort_msg


@@:
%if 0
	call disp_filename
	mov al, 13
	call disp_al
	mov al, 10
	call disp_al
%else
	mov ah, 41h
	lds dx, [di]
	int 21h
%endif
	jmp ..@exit


		; Check for given string (cap-insensitive)
		;
		; INP:	si-> input string to check (either cap),
		;	 terminated by CR (13), semicolon, space, tab,
		;	 comma, colon, [, ], (, or )
		;	dx-> ASCIZ string to check (all-caps)
		; OUT:	Iff string matches,
		;	 ZR
		;	 si-> at separator that terminates the keyword
		;	else,
		;	 NZ
		;	 si = input si
		; STT:	ds = es = ss
		; CHG:	dx, al
isstring?:
	push si
	xchg dx, di
.loop:
	lodsb
	call capitalise
	scasb
	jne .mismatch
	test al, al
	jne .loop
	jmp .matched_zr

.mismatch:
	call iseol?
	je .checkend
	cmp al, 32
	je .checkend
	cmp al, 9
	je .checkend
	cmp al, '.'
	je .checkend
	cmp al, ','
	je .checkend
	cmp al, '='
	je .checkend
	cmp al, ':'
	je .checkend
	cmp al, '['
	je .checkend
	cmp al, ']'
	je .checkend
	cmp al, '('
	je .checkend
	cmp al, ')'
	je .checkend
.ret_nz:
		; NZ
	pop si
.ret:
	xchg dx, di
	retn

.checkend:
	cmp byte [es:di - 1], 0
	jne .ret_nz
.matched_zr:	; ZR
	pop di			; (discard)
	lea si, [si - 1]	; -> separator
	jmp .ret


strlen:
	mov al, 0
	mov cx, -1
	repne scasb
	not cx
	dec cx
	retn


	align 2, db 0
iispentry handler, 0, handler
	call near $
.displacement: equ $ - 2
.displacebase:
.hwreset:
	retf

.number:
	dw 0

	align 2, db 0
	endarea handler


	align 2, db 0
stub:
.entry:
	push cs
	jmp 0:commonhandler
.segment1: equ $ - 2

iispentry i2D, 0, i2D
	cmp ah, 0
.multiplexnumber: equ $ - 1
	je .ours
	align 2, nop
	call near stub.entry
.hwreset:
	retf

.number:
	dw 2Dh

.ours:
	jmp 0:int2Dhandler
stub.segment2: equ $ - 2

	align 2, db 0
	endarea stub


	align 2, db 0
commonhandler:
	lframe 0
	lpar word, iretflags
	lpar word, iretcs
	lpar word, iretip
	lpar word, handleroffset
	lpar word, handlersegment
	push ax
	push bx
	push cx
	push dx
	push es
	push ds
	push si
	push di
	lpar 16, regs
	lenter
	mov si, sp			; -> bp on stack, then di, si, etc
	 push cs
	 pop ds
	lvar word,	options
	mov di, 0			; ! preserve flags
	 push di
	lvar word,	amisversion
	lvar dword,	amispointer
	lvar 16,	amissignature
	lreserve			; reserve space for AMIS data (with lea)
	pushf
	mov dx, word [format]
	and dx, ~ _format_raw
	; test dx, dx			; format v0 ?
	jz @F
	inc word [bp + ?options]	; no, indicate in ?options
@@:
	cld
	mov di, word [bp + ?handleroffset]
	mov es, word [bp + ?handlersegment]
					; es:di -> behind near call
	inc di				; -> at interrupt number word
	mov bx, word [es:di]
	cmp bx, 2Dh			; special ?
	jne @F
	testopt [ctrl0], 40h		; do we want to record AMIS calls ?
	jz .skip			; no, skip -->

		; The following is a hack: If we're running quiet,
		;  then hide AMIS installation checks which result
		;  in al=00h (multiplex number not in use).
		; This greatly reduces the amount of records kept
		;  from our AMIS multiplexer users, which may scan
		;  all 256 multiplex numbers once or twice.
	test al, al			; AMIS installation check hack ?
	jnz @F				; no, record it -->
	testopt [bp + ?options], 1
	jnz .check
	testopt [options], _quiet	; quiet ?
	jz @F				; no, record it -->
.check:
	 popf
	 pushf
	push di				; AMIS func 00h CHGs al, dx, di, cx
	pushf				; make int style stack frame
	call far [es:di - (i2D.number - i2D) + ieNext]
					; call the handler
	cld
	testopt [bp + ?options], 1
	jz .onlypresence
	test al, al			; is in use ?
	jz .onlypresence

		; Additional hack: In v1 format we remember the
		;  signature of AMIS multiplexers that replied.
	mov word [bp + ?amispointer], di
	mov word [bp + ?amispointer + 2], dx
	push es
	push ds
	push si
	mov word [bp + ?amisversion], cx
	 mov ds, dx
	mov si, di			; ds:si -> their signature
	 push ss
	 pop es
	lea di, [bp + ?amissignature]
	mov cx, 16			; (copy bytewise, avoids segment boundary)
	rep movsb			; copy the signature
	pop si
	pop ds
	pop es
	setopt [bp + ?options], 2
.onlypresence:
	pop di				; (restore our pointer)
	testopt [options], _quiet	; quiet ?
	jz @F				; no, record it -->
	test al, al			; is in use ?
	jz .skip			; no, skip it -->
@@:
	testopt [ctrl0], 80h
	jz .skip

	push si
	push bx
	call bx_intnum_to_dh_bx_intsel	; get mask and offset for bit array
	test byte [quietinterrupts + bx], dh
	pop bx
	jz .intend

	mov si, excludeints.first
	mov cx, word [excludeints.inuse]
	jcxz .intend
.intloop:
	cmp bl, byte [si]
	jne .intnext
	test byte [si + 1], eifMatchAL
	jz @F
	cmp al, byte [si + 2]
	jne .intnext
@@:
	test byte [si + 1], eifMatchAH
	jz @F
	cmp ah, byte [si + 3]
	jne .intnext
@@:
	pop si
	jmp .skip
.intnext:
	add si, 4
	loop .intloop
.intend:

	mov ax, word [bp + ?iretcs]
	mov si, excludesegments.first
	mov cx, word [excludesegments.inuse]
	jcxz .segend
.segloop:
	cmp ax, word [si]
	jb .segnext
	cmp ax, word [si + 2]
	ja .segnext
	pop si
	jmp .skip
.segnext:
	add si, 4
	loop .segloop
.segend:
	pop si

	mov ax, word [free]		; ~> first free slot, if any
	cmp ax, word [size]		; any free ?
	jae .skip			; no -->

	mov cx, swi_info_v0_size
	mul cx				; dx:ax = byte offset
	mov cl, 4
	ror dx, cl			; bit 16..23 rotated into high 4 bits
	cmp ax, 8000h			; higher half ?
	jb @F
	add dx, 800h			; address higher half
	sub ax, 8000h			; adjust down offset
		; (This adjustment insures that we do not
		;  straddle a segment boundary at 64 KiB.)
@@:
	add dx, word [swi_info_array_segment]
	push ss
	pop ds				; ds:si -> our stack frame
	mov cx, 9
	 push es
	 push di			; (preserve entrypoint)
	mov es, dx
	mov di, ax			; -> next array entry
	rep movsw			; copy 9 registers (ds, es, 7 gpr)
	add si, 4			; skip what will become far return address
	mov cl, 3
	rep movsw			; copy iret stack frame
	mov ax, bx
	stosw				; interrupt number
	inc word [cs:free]		; count as used

	testopt [bp + ?options], 2
	jz .skip_pop_di_es

	 push cs
	 pop ds
	mov ax, word [free]		; ~> first free slot, if any
	cmp ax, word [size]		; any free ?
	jae .skip_pop_di_es		; no -->

	 push ss
	 pop ds
	lea si, [bp + ?amissignature]
	mov cx, 8 + 2 + 1		; sig, pointer, version
	rep movsw
	xor ax, ax
	stosw				; reserved
	mov ah, 1
	stosw				; 100h = AMIS signature
	inc word [cs:free]		; count as used

.skip_pop_di_es:
	 pop di
	 pop es				; (preserve entrypoint)
.skip:
	cmp bx, 2Dh			; special ?
	jne @F
	sub di, (i2D.number - i2D) - (handler.number - handler)
					; fix up
@@:
		; Jump to the handler stored in our downlink.
	 push word [es:di - (handler.number - handler) + ieNext + 2]
	 push word [es:di - (handler.number - handler) + ieNext]
	 pop word [bp + ?handlersegment]
	 pop word [bp + ?handleroffset]	; ! seg/ofs reversed
	popf
	lleave , forcerestoresp
	pop di
	pop si
	pop ds
	pop es
	pop dx
	pop cx
	pop bx
	pop ax
	retf			; jump to next handler


%if 0

Supported Int2D functions:

AMIS - Installation check
INP:	al = 00h
OUT:	al = 0FFh
	cx = Private version number (currently 0100h)
	dx:di-> signature: "ecm     ","INTERCEP"

AMIS - Get private entry point - NOP: no private entry point
INP:	al = 01h
OUT:	al = 00h

AMIS - Uninstall - NOP: can't uninstall
INP:	al = 02h
OUT:	al = 02h (Cannot uninstall yet, will do so when able)

AMIS - Request pop-up - NOP: no pop-up
INP:	al = 03h
OUT:	al = 00h

AMIS - Determine chained interrupts
INP:	al = 04h
OUT:	al = 04h
	dx:bx-> interrupt hook list (Int2D always.)

AMIS - Get hotkeys - NOP: no hotkeys
INP:	al = 05h
OUT:	al = 00h

AMIS - Get device driver information - NOP: no device
INP:	al = 06h
OUT:	al = 00h

AMIS - Reserved for AMIS
INP:	al = 07h..0Fh
OUT:	al = 00h

TSR - Get compatible version
INP:	al = 10h
OUT:	al = 0FFh
	cx = compatible version number (currently 0100h)

TSR - Get ctrl0
INP:	al = 20h
OUT:	al = size of returned control data (not 0)
	dx:bx-> control data
	Unsupported bits switched off

TSR - Reserved for TSR
INP:	al = 11h..1Fh, 21h..FFh
OUT:	al = 00h

%endif


	align 2, nop
int2Dhandler:
		; This structure must exactly match an IISP header.
	jmp strict short @F	; entry
	dw .actual
.cs:	dw 0			; this far pointer points to .actual entrypoint
	dw 0			; signature
	db 0			; EOI flag
	dw 0			; hw reset entry
	times 7 db 0		; reserved
%if $ - int2Dhandler != IISPENTRY_size
 %error Structure does not match.
%endif
@@:
		; This code must exactly match our i2D entrypoint.
	cmp ah, 0
	je @F
@@:
	align 2, nop
	call near .common
.hwreset:
	retf

.number:
	dw 2Dh
%if $ - int2Dhandler != i2D.ours - i2D
 %error Structure does not match.
%endif
		; End of matching code.

.common:
	push cs			; setup stack as for actual entrypoint
	jmp commonhandler	; jump near to commonhandler

.actual:
	test al, al
	jz .installationcheck		; installation check -->
	cmp al, 02h
	je .iret			; uninstallation -->
	cmp al, 04h
	je .determineinterrupts		; determine hooked interrupts -->
	cmp al, 10h
	je .getver			; get compatible version -->
	cmp al, 20h
	je .getctrl0			; get ctrl0 -->
				; all other functions are reserved or not supported by TSR
.nop:
	mov al, 0			; show not implemented
	iret

.installationcheck:
	dec al				; (= FFh) show we're here
	mov cx, 100h			; = version
	mov di, msg.ouramissig		; dx:di -> AMIS signature strings of this program
.iret_dx_cs:
	mov dx, cs
.iret:
	iret

.determineinterrupts:			; al = 04h, always returns list
	mov bx, word [cs:amisintr]	; dx:bx -> hooked interrupts list
	mov dx, word [cs:amisintr + 2]
	iret

.getver:
	mov al, 0FFh			; show call supported
	mov cx, 100h			; = compatible version
	iret

.getctrl0:
	mov bx, ctrl0			; dx:bx -> ctrl0
	and byte [cs:bx], 0C0h
	mov al, 1			; size of ctrl0: 1 byte
	jmp short .iret_dx_cs


capitalise:
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
	retn


getnybble:
	cmp al, '0'
	jb .invalid		; unknown character -->
	cmp al, '9'
	jbe @F
	call capitalise
	cmp al, 'A'
	jb .invalid
	cmp al, 'F'
	ja .invalid
	sub al, 'A' - 10 - '0'
@@:
	sub al, '0'
	db __TEST_IMM8		; (NC, skip stc)
.invalid:
	stc
	retn


getdecit:
	cmp al, '0'
	jb .invalid		; unknown character -->
	cmp al, '9'
	ja .invalid
	sub al, '0'
	db __TEST_IMM8		; (NC, skip stc)
.invalid:
	stc
	retn


TOUPPER		equ ~32

		; (From lDebug revision hg 1ff99456fa11)
		;
		; Subroutine to process an FCB.
		;
		; INP:	di -> FCB
		;	si -> input
nn6:
	lodsb
	call iseol?.notsemicolon
	je nn7			; if end
	call ifsep
	je nn6			; if separator (other than CR)
	cmp al, byte [switchar]
	je nn10			; if switch character
nn7:
	dec si
	mov ax, 2901h		; parse filename
	int 21h
	push ax			; save AL
nn8:
	lodsb			; skip till separator
	call ifsep
	je nn9			; if separator character (including CR)
	cmp al, byte [swch1]
	jne nn8			; if not switchar (sort of)
nn9:
	dec si
	pop ax			; recover AL
	cmp al, 1
	jne nn9a		; if not 1
	dec ax
nn9a:
	retn

		; Handle a switch (differently).
nn10:	lodsb
	call iseol?.notsemicolon
	je nn7			; if end of string
	call ifsep
	je nn10			; if another separator (other than CR)
	mov al, 0
	stosb
	dec si
	lodsb
	cmp al, 'a'
	jb nn11			; if not a lower case letter
	cmp al, 'z'
	ja nn11
	and al, TOUPPER		; convert to upper case
nn11:	stosb
	mov ax, 32<<8|32
	stosw
	stosw
	stosw
	stosw
	stosw
	xor ax, ax
	stosw
	stosw
	stosw
	stosw
	retn			; return with al = 0


		; Compare character with separators
		;
		; INP:	al = character
		; OUT:	ZR if al is CR, NUL, blank, tab, comma, semicolon, or equal sign
		;	NZ else
		; REM:	This is only used for parsing FCBs.
ifsep:
	call iseol?		; semicolon or CR or NUL
	je .return
	cmp al, 32
	je .return
	cmp al, 9
	je .return
	cmp al, ','
	je .return
	cmp al, '='
.return:
	retn


iseol?:
	cmp al, ';'
	je .ret
.notsemicolon:
	cmp al, 13		; this *IS* iseol?
	je .ret
	cmp al, 0
.ret:
	retn


		; INP:	ds:si -> source IISP header (or pseudo header)
		;	es:di -> destination IISP header
		; OUT:	EI
		;	si and di both incremented by 6
		; CHG:	-
		; STT:	UP
update_iisp_header:
	push ax
	mov ax, word [cs:debuggerfunction]
	test ax, ax			; found the debugger ?
	jz @F				; no -->
	int 2Dh				; call its Update IISP Header function
	cmp al, 0FFh			; supported ?
	pop ax
	je .ret				; yes. done -->
	db __TEST_IMM8			; (skip pop)
@@:
	pop ax				; restore ax, then do manual update
	cli				; try to rest while updating chain
	cmpsw				; skip over first word (entrypoint)
					;  (generally xxEBh or 0EA90h)
	movsw
	movsw				; transfer source ieNext to dest ieNext
	sti
.ret:
	retn


		; INP:	al = interrupt number
		;	ds:si-> interrupt entry
		; OUT:	CY if unhooking failed
		;	NC if unhooking successful
		; CHG:	ah, es, di, si
UnhookInterrupt:
			; UnhookInterruptSim (below) only checks if it's possible to unhook this interrupt.
			; This function really unhooks the interrupt if possible.
			;
			; This is to cover the situation when some of the hooked interrupts can unhook,
			; but some can't. If the uninstaller would start to unhook the interrupts and then
			; catch the interrupt that can't be unhooked the user would end up with a dead TSR
			; that's uninstalled halfway. Very bad.
			;
			; "Simulating" the unhooking first and checking if all interrupts can unhook
			; usually will not return such a state.
	call UnhookInterruptSim
	jc .ret				; bad. --> (CY)
	jz .easy
.hard:
				; "hard" case: UnhookInterruptSim has however already done the work,
				; so the hard case is here indeed easier than the easy case.
	call update_iisp_header		; copies our stored pointer into the other's entry
	clc
	retn
.easy:
	push ds
	push dx
	lds dx, [ si + 2 ]		; get what we stored in the entry
	mov ah, 25h			; easy case - just reset to the value stored
	int 21h				; doesn't alter CF (leaves NC from UnhookInterruptSim) or sets NC
	pop dx
	pop ds
.ret:
	retn

		; INP:	ds:si-> IISP entry
		;	al = interrupt number
		; OUT:	NC if no error (either hard or easy case),
		;	 ZR if easy case,
		;	  ds:si-> our IISP entry, containing stored interrupt
		;	 NZ if hard case,
		;	  ds:si-> our IISP entry
		;	  es:di-> IISP entry to modify
		;	CY if error (not first handler and no IISP chain to this handler)
		; CHG:	ah, es, di
UnhookInterruptSim:
	push bx

	; harden this, check we are an IISP entry
	 push ds
	 pop es				; es => our handler segment
	mov bx, si			; es:bx -> our handler
	call IsIISPEntry?		; does it have an IISP header ?
	jne .fail			; fail if not

	mov ah, 35h			; get current vector
	int 21h				; es:bx-> current interrupt handler
	cmp si, bx			; our pointer ?
	jne .hard

	push ax
	push si
	mov si, ds
	mov ax, es
	 cmp si, ax			; our segment ?
	pop si
	pop ax
	jne .hard

	and ah, 00h			; NC, ZR
	pop bx
	retn
.hard:
		; INP:	ds:si-> IISP entry
		;	es:bx-> current interrupt entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP entry
		;	 es:di-> IISP entry to modify
		; CHG:	ah, es, di, (bx)
	call SearchIISPChain
	jne .harder
.found:				; found reference to our interrupt handler
	mov di, bx			; es:di-> IISP entry that references our's
	or ah, 0FFh			; NC, NZ
	pop bx
	retn

.harder:			; Desperate attempt to find IISP entry that references ours by
				; searching through the interrupts hooked by other AMIS TSRs. Note
				; that the plexer loop will find and search through the list of
				; hooked interrupts of the uninstalling TSR itself, but this causes
				; no trouble.
		; INP:	ds:si-> IISP entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP entry
		;	 es:di-> IISP entry to modify
		; CHG:	ah, es, di, (bx)
	d4bp
	push dx
	push ax				; register with interrupt number last
	xor ax, ax
.loopplex:
	mov al, 00h			; AMIS installation check
	push cx
	int 2Dh				; enquire whether there's anyone
	pop cx				;  but we don't care who it might be
	inc al
	jz .search
.nextplex:
	inc ah
	jnz .loopplex			; try next multiplexer -->
	pop ax
	pop dx
.fail:					; IISP incompatible TSR between current interrupt entry and our entry
					;  and no AMIS compatible TSR installed on top of our entry
	stc
	pop bx
	retn

		; INP:	ah = multiplex number of AMIS TSR to search through
		;	ss:sp-> interrupt number (byte), must be preserved
		; CHG:	es, di, dx, bx
.search:
	mov al, 04h
	pop bx
	push bx				; low byte is the interrupt number
	int 2Dh
	cmp al, 03h			; returned its interrupt entry ?
				; RBIL doesn't explicitly state that this interrupt entry has to
				; be IISP compatible. But I'm too lazy to look up the older AMIS,
				; and SearchIISPChain checks the interrupt entry anyway.
	je .search_dxbx
	cmp al, 04h			; returned list of hooked interrupts ?
	jne .nextplex			; no, try next multiplexer -->
	mov di, bx
	pop bx
	push bx				; bl = interrupt number
	mov al, bl
.search_intlist_seg:
	mov es, dx			; es:di-> list
.search_intlist:		; Search the returned list for the required interrupt number.
	scasb				; our interrupt number ?
	je .search_found_intlist
	cmp byte [es:di-1], 2Dh		; was last in list ?
	je .nextplex
	scasw				; skip pointer
	jmp short .search_intlist	; try next entry -->
.search_found_intlist:
	mov bx, word [es:di]		; dx:bx-> IISP entry
	scasw				; skip pointer
	call SearchIISPChain
	je .search_found		; found entry -->
		; This specific jump supports TSRs that hook the same
		; interrupt more than once; jumping to .nextplex instead
		; (as previously) aborts the search after the first match
		; in the interrupt list. This support might become useful.
	cmp al, 2Dh			; was last in list ?
	je .nextplex
	jmp short .search_intlist_seg

.search_dxbx:
	mov es, dx			; es:bx-> (IISP) interrupt entry
				; The entry we found now is possibly behind the non-IISP entry that
				; terminated our first SearchIISPChain call (at .hard). We then
				; possibly might find our entry in this hidden part of the chain.
	call SearchIISPChain
	jne .nextplex			; didn't find our entry in the chain -->
.search_found:
	pop ax
	pop dx
	jmp short .found


SearchIISPChain.next:
	les bx, [es:bx +2]		; get next interrupt entry

		; INP:	ds:si-> IISP entry
		;	es:bx-> current interrupt entry
		; OUT:	NZ if reference to ds:si not found in IISP chain es:bx->
		;	ZR if reference found,
		;	 es:bx-> IISP (or uninstalled iHPFS) interrupt entry with reference
		; CHG:	es, bx
SearchIISPChain:
	call IsIISPEntry?			; that an IISP entry ?
	jnz .return				; nope --> (NZ)
	cmp si, word [ es:bx + ieNext ]		; our pointer ?
	jne .next				; no, try next -->
	push ax
	mov ax, ds
	cmp ax, word [ es:bx + ieNext + 2]	; our segment ?
	pop ax
	jne .next				; no, try next -->
.return:					; yes, found (ZR)
	retn


		; If unhooking (or hooking) failed after the simulation succeeded for all
		; interrupts, at least one interrupt changed between the simulation and the
		; attempt to really (un)hook that interrupt. This is kind of a critical
		; error because it'll happen very rarely, and I'd be interested to hear
		; about software with which it does happen.
unhookerror:
	mov cx, 16			; 16 words * 16 bits / word = 256 bits
	 push ax
	xor ax, ax			; = 0
	push cs
	pop es
	mov di, selectedinterrupts	; es:di -> interrupts bit array
	rep stosw
	 pop ax
	mov cx, unhookerror_shell
unhookerrorcritical:
	call disp_progname
	mov dx, msg.cantuninstall
.install:
	call disp_msg

	 pop dx
	cmp dx, byte -1			; at least another interrupt failed ? (at least two interrupts failed ?)
	 push dx
	je .single
	mov dx, msg.interrupts		; first message for multiple interrupts
	jmp short .first
.single:
	mov dx, msg.interrupt		; first message for single interrupt
.first:
	mov bx, dx

.loop:
	call disp_msg			; msg.interrupts or msg.hcomma
	call disp_al_hex
	 push cx
	 push bx
	xchg bx, ax			; bx = number
	call bx_intnum_to_dh_bx_intsel
	or byte [cs:selectedinterrupts + bx], dh
					; mark as failed
	 pop bx
	 pop cx
	pop ax				; get next value from stack
	 pop dx
	cmp dx, byte -1			; at least two more ?
	 push dx
	mov dx, msg.hcomma		; message if at least two more interrupts to display
	jne .notlast			; yes -->
	mov dx, msg.hand		; only one more interrupt to display
.notlast:
	cmp ax, byte -1			; last error marker?
	jne .loop			; no, keep looping -->

	mov dx, msg.hooked		; last message for single interrupt
	cmp bx, msg.interrupt		; was first message for single interrupt ?
	je .abort			; yes -->
	mov dx, msg.hookeds		; else use last message for multiple interrupts
.abort:
	jmp cx


		; In this case, no critical error occurred.
		;  Instead of panicking, allow the user to
		;  run a shell or debugger as a user command.
unhookerror_shell:
	pop ax
	pop ax				; discard pointer to AMIS int list
	pop ax				; discard flags return
	call disp_msg
	call disp_progname
	mov dx, msg.shell
	call disp_msg

.loop:
	push cs
	pop ds
	push cs
	pop es

	cmp byte [linein + 1], 0
	je .ldebug

	xor ax, ax
	mov si, word [execblock.tail]
	lea di, [si - 1]
	lodsb				; = length of tail
	inc ax				; include the terminator CR
	xchg cx, ax
	rep movsb			; move tail back

	jmp .prompt

.ldebug:
	mov si, msg.ldebug_shell
	mov cx, msg.ldebug_shell.length
	mov di, linein + 2
	rep movsb

	mov dl, 32
	xor bx, bx
@@:
	push bx
	call bx_intnum_to_dh_bx_intsel
	test byte [selectedinterrupts + bx], dh
	pop bx
	jz @F

	cmp di, linein + 2 + 255 - 2 - 3
	ja @FF
	mov al, dl
	stosb
	mov al, bl
	call hexbyte
	mov dl, ','
@@:
	inc bl
	jnz @BB

@@:
	mov ax, "'" + (13 << 8)
	stosw
	lea ax, [di - 1 - (linein + 2 - buffers + transient_size + 256)]
	mov byte [linein + 1], al

.prompt:
	mov dx, linein
	int3
	mov ah, 0Ah
	int 21h				; read input line
	mov dx, msg.crlf
	call disp_msg
	mov si, linein + 2
@@:
	lodsb
	cmp al, 32
	je @B
	cmp al, 9
	je @B
	call scancl.setupexec
	jnc execute

	call disp_progname
	mov dx, msg.shell.fail
	call disp_msg
	jmp .loop


panic_and_critunins:
	call disp_msg
	mov dx, msg.criticalunins	; else display critical failure message
panic:
	call disp_msg
	call restorestate
	mov dx, msg.panic
	call disp_msg
@@:
	int3
	sti
	hlt
	jmp @B


..@stack equ $ - (($-$$) % 2)		; (above is overwritten by stack while installing)
..@exit:				; terminate successful, re-used by some code (install too)
	call restorestate
	mov ax, 4C00h
	int 21h


hexbyte:
	push cx
	mov cl, 4
	rol al, cl
	call hexnyb
	rol al, cl
	pop cx

hexnyb:
	push ax
	and al, 0Fh
.common:
		; These three instructions change to ASCII hex.
		;  Refer to https://codegolf.stackexchange.com/questions/193793/little-endian-number-to-string-conversion/193842#193842
	cmp al, 10	; set CF according to digit <= 9
	sbb al, 69h	; read CF, set CF and conditionally set AF
	das		; magic, which happens to work
	stosb
	pop ax
	retn


		; INP:	ax = first failed interrupt
		;	ss:sp -> list of words of further failed interrupts
		;		(terminated by FFFFh entry)
		;	behind terminator: stack with dword far -> int list
hookerror:
	mov cx, abort_msg
	; jmp short displayhookerror

		; INP:	cx -> code to jump to after displaying the list
		;		of interrupts that failed to install
		;		(will receive stack emptied of int numbers)
displayhookerror:
	call disp_progname
	mov dx, msg.cantinstall
	call disp_msg

	 pop dx
	cmp dx, byte -1			; at least another interrupt failed ? (at least two interrupts failed ?)
	 push dx
	je .single
	mov dx, msg.interrupts		; first message for multiple interrupts
	jmp short .first
.single:
	mov dx, msg.interrupt		; first message for single interrupt
.first:
	mov bx, dx

.loop:
	call disp_msg			; msg.interrupts or msg.hcomma
	call disp_al_hex
	pop ax				; get next value from stack
	 pop dx
	cmp dx, byte -1			; at least two more ?
	 push dx
	mov dx, msg.hcomma		; message if at least two more interrupts to display
	jne .notlast			; yes -->
	mov dx, msg.hand		; only one more interrupt to display
.notlast:
	cmp ax, byte -1			; last error marker?
	jne .loop			; no, keep looping -->

	mov dx, msg.invalid		; last message for single interrupt
	cmp bx, msg.interrupt		; was first message for single interrupt ?
	je .ret				; yes -->
	mov dx, msg.invalids		; else use last message for multiple interrupts
.ret:
	jmp cx


		; INP:	ds:bx -> AMIS interrupt list
		;	dx = code called for each interrupt (with ds:si-> interrupt entry, al = interrupt number)
		;	cx = code to return to
		; OUT:	CY if any interrupt failed,
		;	 ax = first failed interrupt
		;	 ss:sp-> 0FFFFh terminated list of failed interrupts on stack
		;	 If dx == HookInterrupt,
		;	  bx -> interrupt list entry that failed to hook
		;	NC if all interrupts successfully uninstalled
		; CHG:	si, ax, bx, what called code (dx) changes
		;
		; Called code (dx) has to preserve dx, bx, cx, al.
loopamisintr:

	mov ax, 0FFFFh
	push ax
.loop:
	mov al, byte [ bx ]
	mov si, word [ bx+1 ]

	call dx
	jnc .noerror			; no error -->
	xor ah, ah
	push ax				; else remember number of interrupt, but continue looping

	cmp dx, HookInterrupt
	je .abort			; if during hooking (!!), abort now -->

.noerror:
	add bx, byte 3
	cmp al, 2Dh
	jne .loop			; do until 2Dh done -->

.abort:
	pop ax
	cmp ax, byte -1			; if it's below (CY), there were errors
	jmp cx


		; INP:	bx = interrupt number 0..255
		; OUT:	dh = mask, bx = offset into selectedinterrupts
		; CHG:	cl
bx_intnum_to_dh_bx_intsel:
	mov cl, bl
	and cl, 7
	shr bx, 1
	shr bx, 1
	shr bx, 1
	mov dh, 1
	shl dh, cl
	retn


abort_msg:
	call disp_msg
abort:
	call restorestate
	mov ax, 4CFFh
	int 21h


		; Advanced Critical Failure Management(TM)
		;
		; In case of a critical installation failure (see docs and commentary),
		; we'll try to unhook all the interrupts that we successfully hooked.
		; If that succeeded, the whole TSR can simply be freed again and the
		; problem of the partway installed TSR has been nullified. This is still
		; considered a case to report, but isn't critical in itself because
		; there's no TSR lingering around.
handle_crithookfailure:
	mov si, bx		; save address of failed interrupt's list entry
	mov cx, .critintdisplayed
	jmp displayhookerror
.critintdisplayed:
	call disp_msg		; display the last message (displayhookerror expects
				; to jump to abort_msg)
	mov cx, si		; restore address (don't use the stack here)

		; We now don't simulate the unhooking because we just want to unhook
		; these interrupts that will unhook no matter what. Additionally, it's
		; very likely that all the unhooking succeeds.
		;
		; This code uses an unrolled form of loopamisintr copied from there
		; because I just can't be bothered to add special checking for a
		; shorter list there. I also considered just calling loopamisintr
		; then filtering out the errors of the interrupts that we didn't hook,
		; but that's somewhat hacky too. So copied the code be.
	lds bx, [cs:amisintr]
	mov ax, 0FFFFh
	cmp bx, cx
	je .nullifiedcrit	; the very first interrupt failed to hook, none to unhook -->
	push ax
.lai_loop:
	mov al, byte [ bx ]
	mov si, word [ bx+1 ]

	call UnhookInterrupt
	jnc .lai_noerror	; no error -->
	xor ah, ah
	push ax			; else remember number of interrupt, but continue looping
.lai_noerror:
	add bx, byte 3
	cmp bx, cx
	jne .lai_loop		; do until the next one would be the one that didn't hook -->

	pop ax
	cmp ax, byte -1		; if it's below (CY), there were errors
	jc .critical		; there were errors -->

		; One of the interrupts critically failed to hook, but we were able to
		; unhook everything again.
.nullifiedcrit:
	mov dx, msg.criticalundone
	jmp abort_msg

		; One of the interrupts critically failed to hook. Additionally, at
		; least one other interrupt failed to unhook.
.critical:
	mov dx, msg.criticalins
	mov cx, panic
	jmp unhookerrorcritical.install	; this TSR be damned -->


		; This one works similar to UnhookInterruptSim. Here we check that
		; the interrupt currently contains a valid address since we want to
		; chain to the previous handler.
		;
		; INP:	ds:si-> IISP entry
		;	al = interrupt number
		; OUT:	NC if current interrupt handler valid, or we install a non-chaining handler
		;	CY if current interrupt handler invalid (offset FFFFh or segment 0000h)
		; CHG:	ah
HookInterruptSim:
	push es
	push bx

	; harden this, check we are an IISP entry
	 push ds
	 pop es				; es => our handler segment
	mov bx, si			; es:bx -> our handler
	call IsIISPEntry?		; does it have an IISP header ?
	jne .error			; fail if not

	test byte [ si + ieEOI ], 80h		; should be last handler ?
	jnz .valid				; yes, handler doesn't matter if we don't chain --> (NC)
	mov ah, 35h
	int 21h
	inc bx					; offset FFFFh ?
	jz .error
	mov bx, es
	test bx, bx				; or segment 0000h ?
	jnz .valid				; no, valid --> (NC)
.error:
	stc					; invalid, CY
.valid:
	pop bx
	pop es
	retn


		; INP:	ds:si-> IISP entry
		;	al = interrupt number
		; OUT:	NC if current interrupt handler valid, or we install a non-chaining handler
		;	CY if current interrupt handler invalid (offset FFFFh or segment 0000h),
		;	 though this SHOULD never happen
		; CHG:	ah, es
HookInterrupt:
	call HookInterruptSim
	jc .ret
	push bx
	push dx
	mov ah, 35h
	int 21h					; get current handler
	test byte [ si + ieEOI ], 80h		; should be last handler ?
	jz .notlast
		; If we are a last (non-chaining) handler
		;  then HookInterruptSim always shortcuts
		;  to succeeding. However, the family of
		;  IsIISPEntry? functions doesn't check
		;  for invalid handler addresses the same
		;  way as HookInterruptSim. (An offset of
		;  FFFFh happens to be rejected by both,
		;  but that is not true of a zero segment.)
		; If we were to install a chaining handler
		;  then the invalid address would be reason
		;  not to install. Consequently, when we
		;  install a last handler then it should
		;  not be installed into an IISP chain that
		;  is reachable from the first handler,
		;  the address of which was invalid.
		; Instead it should install atop the handler
		;  that has an invalid address. Therefore,
		;  check for a segment of zero here.
		; (This check could be added to the function
		;  IsIISPEntry? but it may be useful to have
		;  them otherwise allow addresses with a
		;  segment of zero.)
	push ax
	mov ax, es				; get the segment
	test ax, ax				; is it zero ?
	pop ax
	jz .notlast				; yes -->
	call IsChainingIISPEntry?		; is the first handler a chaining IISP ?
	jne .notlast				; no, can't hook below it -->
	db __TEST_IMM16				; (skip 2 pop ax)
.loop:
	pop ax
	pop ax					; discard last entry's address
	push es
	push bx
	les bx, [ es:bx + ieNext ]		; load next entry
	call IsChainingIISPEntry?		; this one also a chaining IISP ?
	je .loop
		; This write need not be atomic because
		;  the handler is not yet in the chain.
	mov word [ si + ieNext ], bx
	mov word [ si + ieNext + 2 ], es	; store address of next entry (non-IISP or non-chaining)
	pop bx
	pop es
		; The next write needs to be atomic,
		;  so build a header on the stack and
		;  call the update function here.
	push di
	mov dx, "NH"				; "New Handler"
	push dx					; fake IISP ieSignature
	push ds
	push si					; fake IISP ieNext
	mov dx, 0FEEBh				; "jmp short $"
	push dx					; fake IISP ieEntry
	 push ss
	 pop ds
	mov si, sp				; ds:si -> fake IISP header
	mov di, bx				; es:di -> IISP header to update
	call update_iisp_header			; store our handler's address in this chaining IISP entry
	pop dx					; discard
	pop si
	pop ds					; restore
	pop dx					; discard
	pop di					; restore
	jmp short .return

.notlast:
		; This write need not be atomic because
		;  only once we call 21.25 the handler
		;  is linked into the chain.
	mov word [ si + ieNext ], bx
	mov word [ si + ieNext + 2 ], es	; store
	mov ah, 25h
	mov dx, si				; ds:dx-> interrupt entry
	int 21h					; set new handler
.return:
	pop dx
	pop bx
	clc
.ret:
	retn


		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry,
		;	 or IISP entry that doesn't chain
		;	ZR if IISP entry that chains
IsChainingIISPEntry?:
	test byte [ es:bx + ieEOI ], 80h	; this one a non-chaining handler ? (or non-IISP)
	jnz IsIISPEntry?.return			; yes -->
						; otherwise fall through to check if really an IISP entry

		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry
		;	ZR if IISP entry
IsIISPEntry?:
	cmp bx, - (ieSignature + 2)		; may access word at offset FFFFh ?
	ja .return				; yes, avoid --> (NZ)
	cmp word [ es:bx + ieSignature ], "KB"	; "KB"/424Bh ? ("BK" in MASM)
	jne .return
	cmp word [ es:bx + ieEntry ], 0EA90h	; nop\jmp far imm16:imm16 ?
	je .return				; unused IISP entry (created by iHPFS) -->
	cmp byte [ es:bx + ieEntry ], 0EBh	; jmp short ... ?
		; (This opcode should strictly be jmp short $+18 but there's programs
		; that save an additional jmp opcode by jumping directly into their
		; code even though it's not right behind the header.)
	jne .return
	cmp byte [ es:bx + ieJmphwreset ], 0EBh	; jmp short ... ?
	je .return				; usual IISP entry -->
	cmp byte [ es:bx + ieJmphwreset ], 0CBh	; retf ?
	je .return				; a shorter variant -->
	cmp byte [ es:bx + ieJmphwreset ], 0CFh	; iret ?
.return:
	retn


			; The DOS memory allocation strategy classes are actually "Low then high",
			; "High only" and "High then low". RBIL tells trash regarding that. We've
			; to disable the UMB link to insure only low memory is allocated. Normally,
			; the UMB link is already disabled when the program starts. But some
			; programs, like CMDEDIT 3.21, are buggy and leave the UMB link enabled
			; after they used UMBs.
		; INP:	bx = UMB link status to use
		;	dx = memory allocation strategy to use
		; OUT:	CF/ax error status
setstrat:
	mov ax, 5803h
	int 21h			; set UMB link
	jnc .linkdone		; no error
	cmp ax, byte 1		; "invalid function" ?
	stc
	jne .return		; no, actual error -->
	test bx, bx		; wanted to disable anyway ?
	stc
	jnz .return		; no, regard as error -->
.linkdone:
	mov ax, 5801h
	mov bx, dx
	int 21h			; set strategy
.return:
	retn


i24:
	mov al, 3			; always return fail, to handle the error as a soft one
	iret

i23:
	mov word [ cs:$ ], (__JMP_REL8|__REL16__(.return)<<8)	; don't reenter
	call restorestate
.return:
	stc				; always abort program (what default DOS handler also does)
retf_inst:
	retf

		; Restore modified DOS data
		;
		; CHG:	-
		; USE:	Int21
restorestate:
	push ax
	push bx
	mov bx, word [ cs:restored.strategy ]
	mov ax, 5801h
	int 21h				; restore strategy

	mov bx, word [ cs:restored.umblink ]
	mov ax, 5803h
	int 21h				; restore UMB link
	pop bx
	pop ax
	retn


disp_progname:
	push ax
	push cx
	push ds
	lds dx, [cs:progname]
	mov cx, word [cs:progname.length]
	mov bx, 1
	mov ah, 40h
	int 21h
	pop ds
	pop cx
	pop ax
	retn


disp_filename:
	push ds
	push es
	push di
	push ax
	lds dx, [cs:di]
	les di, [cs:di]
	call strlen
	mov bx, 1
	mov ah, 40h
	int 21h
	pop ax
	pop di
	pop es
	pop ds
	retn


disp_amisnum:
	call disp_al_hex
	mov dx, msg.multiplex.2

disp_msg:
	 push ds
	 push ax
	push cs
	pop ds
	mov ah, 09h
	int 21h
	 pop ax
	 pop ds
	retn


disp_al:
	push dx
	push ax
		mov dl, al
		mov ah, 02h
		int 21h
	pop ax
	pop dx
	retn


disp_ax_hex:
	xchg al, ah
	call disp_al_hex
	xchg al, ah
disp_al_hex:
		push cx
		mov cl, 4
		ror al, cl
		call .nibble
		ror al, cl
		pop cx
.nibble:
		push ax
		and al, 0Fh
		add al, '0'
		cmp al, '9'
		jbe .isdigit
		add al, 'A'-('9'+1)
.isdigit:
		xchg dl, al
		mov ah, 02h
		int 21h
		pop ax
		retn


	; Following call: Display number in ax decimal
	; all registers preserved except dx
disp_ax_dec:			; ax (no leading zeros)
	; In: number in ax
	; Out: displayed
		push bx
		xor bx, bx
.pushax:
		push ax
.pushend:
		or bl, bl
		jz .nobl
		sub bl, 5
		neg bl
.nobl:
		push cx
		mov cx, 10000
		call .divide_out
		mov cx, 1000
		call .divide_out
		mov cx, 100
		call .divide_out
		mov cl, 10
		call .divide_out
							; (Divisor 1 is useless)
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h
		pop cx
		pop ax
		pop bx					; Caller's register
		retn

.divide_out:
	; In: ax = number
	;     cx = divisor
	; Out: ax = remainder of operation
	;      result displayed
		push dx
		xor dx, dx
		div cx				; 0:ax / cx
		push dx				; remainder
		dec bl
		jnz .nobl2
		or bh, 1
.nobl2:
		or bh, al
		jz .leadingzero
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h				; display result
 .leadingzero:
		pop ax				; remainder
		pop dx
		retn


disp_dos_error:
	call disp_msg
	mov dx, msg.dos_error_msg_1
	call disp_msg
	call disp_ax_hex
	mov dx, msg.dos_error_msg_2
	call disp_msg
	mov dx, msg.dos_error_unknown
	cmp ax, dos_error_table.after_last
	jae @F
	mov bx, ax
	shl bx, 1
	mov dx, word [cs:dos_error_table + bx]
@@:
	call disp_msg
	mov dx, msg.dos_error_msg_3
	jmp disp_msg


	align 2, db 0
firstumcb:	dw 9FFFh			; guess of where UMBs start (A000h+)
trymultiplexnumber:
		dw -1		; low byte = 0 if set
				; high byte = multiplex number to try first
debuggerfunction:
		dw 0		; = 0 if unused
restored:
.umblink:	dw 0
.strategy:	dw 0
switchar:	db 0		; switch character
swch1:		db ' '		; switch character if it's a slash
ctrl0:		db 0		; control byte
				; 80h = general toggle, no recording if clear
				; 40h = no recording of AMIS int 2Dh calls if clear
	align 2, db 0
size:		dw _MAX_INTERRUPTS
free:		dw 0
format:		dw _format_v0
int_amount:	dw 0
amisintr:	dw 0		; used as doubleword far pointer
int_segment:	dw 0		; must be immediately after amisintr
swi_info_array_segment:
		dw 0
int_end:	dw 0
linenumber:	dw 1

	numdef EXCLUDESEGMENTS_AMOUNT, 2
	align 4, db 0
excludesegments:
.amount:	dw _EXCLUDESEGMENTS_AMOUNT
.inuse:		dw 2
	; format: low word is lower bound, high word is higher bound
.first:		dw 0
.cs:		dw 0
.umcb:		dw 0
		dw -1
		times _EXCLUDESEGMENTS_AMOUNT * 4 - ($ - .first) db 0

	numdef EXCLUDEINTS_AMOUNT, 16
	align 4, db 0
excludeints:
.amount:	dw _EXCLUDEINTS_AMOUNT
.inuse:		dw 0
	; format: low byte is int match number, next byte is flags,
	;	high word is ax match number
.first:		times _EXCLUDEINTS_AMOUNT * 4 - ($ - .first) db 0

eifMatchAL	equ 1
eifMatchAH	equ 2

outputfilename:	dd msg.defaultoutputfilename
datafilename:	dd msg.defaultdatafilename
temppathname:	dd msg.defaulttemppathname
filterfilename:	dd msg.defaultfilterfilename
execname:	dd 0
progname:	dd msg.defaultprogname
.length:	dw msg.defaultprogname.length
execblock:
.env:		dw 0				; 0 = use parent's env
.tail:		dd 0
.fcb1:		dd 5Ch
.fcb2:		dd 6Ch
aftertempfilename:
		dw 0
options:	dw 0
handle:		dw -1


	align 2, db 0
parameters:			; Valid switches on the command line, followed
				; by a byte with one option flag set.
	db "S", _size
	db "O", _outputfilename
	db "D", _datafilename
	db "L", _longmode
	db "T", _temppathname
	db "P", _filterfilename
	db "A", _amis
	db "Q", _quiet
	db "H", 0
	db "?", 0
	db 0FFh			; table end marker

invalidinterrupts:
	db 1Dh, 1Eh, 1Fh, 22h, 23h, 24h, 30h, 31h
	endarea invalidinterrupts


	align 16, db 0
selectedinterrupts:
	times 32 db 0		; 32 bytes * 8 bit/byte = 256 bits

	align 16, db 0
quietinterrupts:
	times 32 db 0		; 32 bytes * 8 bit/byte = 256 bits


msg:
.debuggeramissig:
	fill 8,32,db "ecm"	; vendor
	fill 8,32,db "lDebug"	; product
.ouramissig:
	fill 8, 32, db "ecm"
	fill 8, 32, db "INTERCEP"
.desc:
	asciz "Monitor software interrupt calls."
%if ($ - .desc) > 64
 %error AMIS description too long
%endif

.defaulttemppathname:
	asciz "."
.defaultoutputfilename:
	asciz "intercep.out"
.defaultdatafilename:
	asciz "interpre.dat"
.tmpdireq:
	db "TMPDIR="
.tmpdireq.length equ $ - .tmpdireq
.tmpeq:
	db "TMP="
.tmpeq.length equ $ - .tmpeq
.tempeq:
	db "TEMP="
.tempeq.length equ $ - .tempeq
.v0:
	asciz "V0"
.v0raw:
	asciz "V0RAW"
.v1:
	asciz "V1"
.v1raw:
	asciz "V1RAW"
.raw:
	asciz "RAW"
.none:
	asciz "NONE"
.defaultprogname:
	db "INTERCEP"
.defaultprogname.length: equ $ - .defaultprogname
.defaultfilterfilename:
	asciz "interpre.exe"
.invalidparameter:
	ascic ": Invalid parameter!",13,10
.invalidswitch.1:
	ascic ": Unknown option ",34
.invalidswitch.2:
.error_memory:
	ascic ": Out of memory!",13,10
.noamisnumber:
	ascicline ": No free AMIS multiplex number!"
.error_create.2:
.error_create_raw.2:
.recording.3.notraw:
.couldntfilter.2:
	ascicline 34
.recording.3.raw:
	ascicline 34," (raw)"
.error_opening.1:
	ascic ": Error during open of ",34
.error_opening.2:
.error_reading.2:
	ascic 34,", "
.error_reading.1:
	ascic ": Error during read of ",34
.invalid_interrupt_1:
	db ": Error, interrupt ",36
.invalid_interrupt_2:
	db "h is invalid to specify.",13,10,36
.internalerror:
	ascicline "INTERCEP: Internal error!"
.templatetoomanyquiet.1:ascic ": Too many quiet template entries in file ",34
.templateinvalid.1:	ascic ": Invalid template entry in file ",34
.templatetoomanyquiet.2:
.templateinvalid.2:	ascic 34," line "
.templatetoomanyquiet.3:
.templateinvalid.3:	ascic "!",13,10
.templatedone.1:	ascic ": Read template file ",34
.templatedone.2:	ascic 34,", "
.templatedone.3:	ascic " lines.",13,10
.error_create.1:ascic ": Cannot open intermediate file in ",34
.error_create_raw.1:ascic ": Cannot create output file ",34
.writeerror.1:	ascic ": Write error on file ",34
.writeerror.2:	ascic 34,", "
.exec_error_user:	ascic ": Error executing user program, "
.exec_error_filter:	ascic ": Error executing filter program, "
.error_cmdtail_too_long:
	ascic ": Error, too long command line tail for filter program.",13,10
.filterreturn.1:
	db ": Filter program returned nonzero return code: ",36
.couldntfilter.1:
	db ": Couldn't run output filter program ",34,36
.recording.1:	db ": Recording up to ",36
.recording.2:	db " SWI records to file ",34,36
.hooking.1:	ascic ": Hooking "
.hooking.2:	ascic " interrupt handlers, on multiplex "
.terminated.1:	ascic ": User program terminated with code "
.filterreturn.2:
.terminated.2:	ascic "h ("
.filterreturn.3:
.terminated.3:	ascicline ")."
.usedtotal.1:	ascic ": Amount records is "
.usedtotal.2:	ascic " used, "
.usedtotal.3:	ascic " total.",13,10
.shell:		ascic ": Enter pathname and command line of program to shell to:",13,10,"> "
.shell.fail:	ascic ": Invalid pathname! Enter new pathname and command line:",13,10,"> "
.ldebug_shell:	db "c:\ldebug.com /c='diml in"
.ldebug_shell.length: equ $ - .ldebug_shell
.help.1:
	db "Usage: ",36
.help.2:
	db " [-l] [-t tmpdir] [-s maxcalls] [-o outfile] [-d datafile]",13,10
	db 9,"[-q] [-a] [-e NONE] [-p filterprogram] [-f V1] [-x number]",13,10
	db 9,"program [args...]",13,10
	db 9,"-l",9,"sets long format output: explanation AND register values",13,10
	db 9,"-t",9,"sets temporary directory for intermediate file to ",34,"tmpdir",34,13,10
	db 9,9,     "(will use TMP or TMPDIR environment vars. if found otherwise)",13,10
	db 9,"-s",9,"sets the maximum number of SWI records to ",34,"maxcalls",34,13,10
	db 9,"-o",9,"names the output filename to ",34,"outfile",34," rather than",13,10
	db 9,9,     "the default name (",34,"intercep.out",34,")",13,10
	db 9,"-d",9,"sets the interrupt data file to use, defaults to INTERPRE.DAT",13,10
	db 9,"-q",9,"quiet mode: do not record functions marked with '?'",13,10
	db 9,"-a",9,"enable recording of AMIS functions (int 2Dh)",13,10
	db 9,"-e NONE",9,"enable recording of calls from all segments",13,10
	db 9,"-p",9,"specify filter program name to call (if not raw format)",13,10
	db 9,"-f",9,"specify format: v0, v0raw, v1, or v1raw",13,10
	db 9,"-x",9,"specify AMIS multiplex number to try first",13,10
	db 9,"program",9,9,"is the name of the program to monitor",13,10
	db 9,"args",9,9,   "are any command-line arguments to be passed",13,10
	db 9,9,9,          "to the monitored program."
.crlf:	db 13,10
	ascic

.logo:	db "INTERCEPT -- monitor DOS and BIOS calls.",13,10
	db 9,"By Ned Konz, 1987",13,10
	db 9,"Datafile option and parsing added by Russell Nelson, 1989",13,10
	db 9,"Ported to NASM and extended by C. Masloch, 2021",13,10
	ascic

.multiplex.2:	db "h.",13,10,36
.panic:		db "PANIC! Cannot continue! Advise to reboot.",13,10,36
.cantuninstall:	db ": Cannot remove",36
.criticalunins:	db "CRITICAL removal failure (report!).",13,10,36
.criticalundone:db "Critical installation failure nullified (report!), "
		db "program fully removed.",13,10,36
.criticalins:	db "CRITICAL installation failure (report!)",36
.cantinstall:	db ": Cannot install",36
.interrupts:	db ", interrupts ",36
.interrupt:	db ", interrupt ",36
.hand:		db "h and ",36
.hcomma:	db "h"
.comma:		db  ", ",36
.hookeds:
.hooked:	db "h hooked AMIS incompatible.",13,10,36
.invalids:
.invalid:	db "h invalid.",13,10,36

.dos_error_msg_1:	ascic "DOS error "
.dos_error_msg_2:	ascic "h ("
.dos_error_msg_3:	ascic ").",13,10

.dos_error_0:		ascic "No error"
.dos_error_1:		ascic "Invalid function"
.dos_error_2:		ascic "File not found"
.dos_error_3:		ascic "Path not found"
.dos_error_4:		ascic "No handles available"
.dos_error_5:		ascic "Access denied"
.dos_error_6:		ascic "Invalid handle"
.dos_error_7:		ascic "MCB destroyed"
.dos_error_8:		ascic "Out of memory"
.dos_error_9:		ascic "MCB invalid"
.dos_error_10:		ascic "Environment invalid"
.dos_error_11:		ascic "Format invalid"
.dos_error_12:		ascic "Access code invalid"
.dos_error_13:		ascic "Data invalid"
.dos_error_14:		ascic "Fixup overflow"
.dos_error_15:		ascic "Invalid drive"
.dos_error_16:		ascic "Attempted to remove current directory"
.dos_error_17:		ascic "Not same device"
.dos_error_18:		ascic "No more files"
.dos_error_19:		ascic "Disk write protected"
.dos_error_20:		ascic "Unknown unit"
.dos_error_21:		ascic "Drive not ready"
.dos_error_22:		ascic "Unknown command"
.dos_error_23:		ascic "Data error"
.dos_error_24:		ascic "Bad request structure length"
.dos_error_25:		ascic "Seek error"
.dos_error_26:		ascic "Unknown media type"
.dos_error_27:		ascic "Sector not found"
.dos_error_28:		ascic "Printer out of paper"
.dos_error_29:		ascic "Write fault"
.dos_error_30:		ascic "Read fault"
.dos_error_31:		ascic "General failure"
.dos_error_32:		ascic "Sharing violation"
.dos_error_33:		ascic "Lock violation"
.dos_error_34:		ascic "Disk change invalid"
.dos_error_35:		ascic "FCB unavailable"
.dos_error_36:		ascic "Sharing buffer overflow"
.dos_error_37:		ascic "Code page mismatch"
.dos_error_38:		ascic "Cannot complete file operation"
.dos_error_39:		ascic "Insufficient disk space"
.dos_error_unknown:	ascic "Unknown error"


	align 2
dos_error_table:
.:
%assign ii 0
%rep 40
	_autodigitsstrdef NUMBER, ii
 %deftok _NUMBER _NUMBER
	dw msg.dos_error_ %+ _NUMBER
 %assign ii ii + 1
%endrep
.after_last:	equ ($ - .) >> 1


	endarea msg

	align 16, db 0
	endarea transient

	section BUFFERS follows=TRANSIENT nobits
buffers:

linein:		resb 256 + 4
.end:
stacktop_without_buffer: equ $ + 512

buffer_size equ 8192
buffer:		resb buffer_size
.end:
stacktop_with_buffer: equ $ + 512
